import pytest


@pytest.fixture(scope='module')
def client_app(test_server):
    from terracotta.client.flask_api import create_app
    yield create_app(test_server)


@pytest.fixture()
def client(client_app, test_server):
    with client_app.test_client() as client:
        yield client


def test_get_app(client, test_server):
    rv = client.get('/')
    assert rv.status_code == 200
    assert 'Terracotta Preview' in rv.data.decode('utf-8')



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
import re
from terracotta import __version__

project = 'Terracotta'
copyright = '2018-2021, the Terracotta contributors'
author = 'Dion Häfner, Philip Graae'

# The short X.Y version
version = re.match(r'(\d+\.\d+\.\d+)', __version__).group(1)
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.intersphinx',
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_typehints',
    'sphinx_click.ext',
]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'friendly'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Extension settings --------------------------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'description': 'A light-weight, versatile XYZ tile server built with Flask and Rasterio',
    'code_font_family': "'Roboto Mono', 'Consolas', 'Menlo', 'Deja Vu Sans Mono', "
                        "'Bitstream Vera Sans Mono', monospace",
    'font_family': "'Lato', Arial, sans-serif",
    'head_font_family': "'Lato', Arial, sans-serif",
    'body_text': '#000',
    'sidebar_header': '#4B4032',
    'sidebar_text': '#49443E',
    'github_banner': 'true',
    'github_user': 'DHI-GRAS',
    'github_repo': 'terracotta',
    'github_button': 'true',
    'github_type': 'star',
    'travis_button': 'true',
    'codecov_button': 'true',
    'logo': 'logo.svg'
}

html_favicon = '_static/favicon.ico'


# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Terracottadoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Terracotta.tex', 'Terracotta Documentation',
     'Dion Häfner, Philip Graae', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'terracotta', 'Terracotta Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'Terracotta', 'Terracotta Documentation',
     author, 'Terracotta', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Custom exec directive ---------------------------------------------------

import sys
from os.path import basename

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from docutils import nodes, statemachine
from docutils.parsers.rst import Directive


class ExecDirective(Directive):
    """Execute the specified python code and insert the output into the document"""
    has_content = True

    def run(self):
        old_stdout, sys.stdout = sys.stdout, StringIO()

        tab_width = self.options.get('tab-width', self.state.document.settings.tab_width)
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)

        try:
            exec('\n'.join(self.content))
            text = sys.stdout.getvalue()
            lines = statemachine.string2lines(text, tab_width, convert_whitespace=True)
            self.state_machine.insert_input(lines, source)
            return []
        except Exception:
            warn_text = f'Unable to execute python code at {basename(source)}:{self.lineno}'
            warning = self.state_machine.reporter.warning(warn_text)
            return [
                warning,
                nodes.error(
                    None,
                    nodes.paragraph(text=warn_text),
                    nodes.paragraph(text=str(sys.exc_info()[1]))
                )
            ]
        finally:
            sys.stdout = old_stdout


# -- Setup function ----------------------------------------------------------

def setup(app):
    app.add_css_file('https://fonts.googleapis.com/css?family=Lato|Roboto+Mono')
    app.add_directive('exec', ExecDirective)



import logging


def test_logstream(caplog, capsys):
    from terracotta import logs
    assert logs.use_colors

    logger = logs.set_logger('DEBUG')
    caplog.set_level('DEBUG', logger='terracotta')
    logger.warning('test')

    log_out = caplog.records
    assert len(log_out) == 1
    log_out = log_out[0]
    assert 'test' == log_out.message
    assert logs.LEVEL_PREFIX['WARNING'] == log_out.levelshortname

    captured = capsys.readouterr()
    assert '[!]' in captured.err
    assert 'test' in captured.err


def test_logstream_nocolors(monkeypatch, caplog, capsys):
    with monkeypatch.context() as m:
        from terracotta import logs
        m.setattr(logs, 'use_colors', False)

        assert not logs.use_colors

        logger = logs.set_logger('DEBUG')
        caplog.set_level('DEBUG', logger='terracotta')
        logger.warning('test')

        log_out = caplog.records
        assert len(log_out) == 1
        log_out = log_out[0]
        assert 'test' == log_out.message
        assert logs.LEVEL_PREFIX['WARNING'] == log_out.levelshortname

        captured = capsys.readouterr()
        assert ' [!] test\n' == captured.err


def test_double_init(caplog):
    from terracotta import logs

    logs.set_logger('DEBUG')
    logs.set_logger('DEBUG')

    caplog.set_level('DEBUG', logger='terracotta')
    logger = logging.getLogger('terracotta')
    logger.warning('test')

    assert len(caplog.records) == 1



import numpy as np
from PIL import Image

import pytest


def test_colormap_handler():
    from terracotta.handlers import colormap
    cmap = colormap.colormap(colormap='jet', stretch_range=[0., 1.], num_values=50)
    assert cmap
    assert len(cmap) == 50
    assert len(cmap[0]['rgba']) == 4
    assert cmap[0]['value'] == 0. and cmap[-1]['value'] == 1.


@pytest.mark.parametrize('stretch_range', [[0, 20000], [20000, 30000], [-50000, 50000]])
@pytest.mark.parametrize('cmap_name', [None, 'jet'])
def test_colormap_consistency(use_testdb, testdb, raster_file_xyz,
                              stretch_range, cmap_name):
    """Test consistency between /colormap and images returned by /singleband"""
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import singleband, colormap

    ds_keys = ['val21', 'x', 'val22']

    # get image with applied stretch and colormap
    raw_img = singleband.singleband(ds_keys, raster_file_xyz, stretch_range=stretch_range,
                                    colormap=cmap_name)
    img_data = np.asarray(Image.open(raw_img).convert('RGBA'))

    # get raw data to compare to
    driver = terracotta.get_driver(testdb)

    with driver.connect():
        tile_data = get_tile_data(driver, ds_keys, tile_xyz=raster_file_xyz,
                                  tile_size=img_data.shape[:2])

    # make sure all pixel values are included in colormap
    num_values = stretch_range[1] - stretch_range[0] + 1

    # get colormap for given stretch
    cmap = colormap.colormap(colormap=cmap_name, stretch_range=stretch_range,
                             num_values=num_values)
    cmap = dict(row.values() for row in cmap)

    # test nodata
    nodata_mask = tile_data.mask
    assert np.all(img_data[nodata_mask, -1] == 0)

    # test clipping
    below_mask = tile_data < stretch_range[0]
    assert np.all(img_data[below_mask & ~nodata_mask, :-1] == cmap[stretch_range[0]][:-1])

    above_mask = tile_data > stretch_range[1]
    assert np.all(img_data[above_mask & ~nodata_mask, :-1] == cmap[stretch_range[1]][:-1])

    # test values inside stretch_range
    values_to_test = np.unique(tile_data.compressed())
    values_to_test = values_to_test[(values_to_test >= stretch_range[0])
                                    & (values_to_test <= stretch_range[1])]

    for val in values_to_test:
        rgba = cmap[val]
        assert np.all(img_data[tile_data == val] == rgba)



from io import BytesIO
import json
import urllib.parse
from collections import OrderedDict

from PIL import Image
import numpy as np

import pytest


@pytest.fixture(scope='module')
def flask_app():
    from terracotta.server import create_app
    return create_app()


@pytest.fixture(scope='module')
def client(flask_app):
    with flask_app.test_client() as client:
        yield client


def test_get_keys(client, use_testdb):
    rv = client.get('/keys')

    expected_response = [
        {'key': 'key1'},
        {'key': 'akey'},
        {'key': 'key2', 'description': 'key2'}
    ]
    assert rv.status_code == 200
    assert expected_response == json.loads(rv.data)['keys']


def test_get_metadata(client, use_testdb):
    rv = client.get('/metadata/val11/x/val12/')
    assert rv.status_code == 200
    assert ['extra_data'] == json.loads(rv.data)['metadata']


def test_get_metadata_nonexisting(client, use_testdb):
    rv = client.get('/metadata/val11/x/NONEXISTING/')
    assert rv.status_code == 404


def test_get_datasets(client, use_testdb):
    rv = client.get('/datasets')
    assert rv.status_code == 200
    datasets = json.loads(rv.data, object_pairs_hook=OrderedDict)['datasets']
    assert len(datasets) == 4
    assert OrderedDict([('key1', 'val11'), ('akey', 'x'), ('key2', 'val12')]) in datasets


def test_get_datasets_pagination(client, use_testdb):
    # no page (implicit 0)
    rv = client.get('/datasets?limit=2')
    assert rv.status_code == 200
    response = json.loads(rv.data, object_pairs_hook=OrderedDict)
    assert response['limit'] == 2
    assert response['page'] == 0

    first_datasets = response['datasets']
    assert len(first_datasets) == 2
    assert OrderedDict([('key1', 'val11'), ('akey', 'x'), ('key2', 'val12')]) in first_datasets

    # second page
    rv = client.get('/datasets?limit=2&page=1')
    assert rv.status_code == 200
    response = json.loads(rv.data, object_pairs_hook=OrderedDict)
    assert response['limit'] == 2
    assert response['page'] == 1

    last_datasets = response['datasets']
    assert len(last_datasets) == 2
    assert OrderedDict([('key1', 'val11'), ('akey', 'x'), ('key2', 'val12')]) not in last_datasets

    # page out of range
    rv = client.get('/datasets?limit=2&page=1000')
    assert rv.status_code == 200
    assert not json.loads(rv.data)['datasets']

    # invalid page
    rv = client.get('/datasets?page=-1')
    assert rv.status_code == 400

    # invalid limit
    rv = client.get('/datasets?limit=-1')
    assert rv.status_code == 400


def test_get_datasets_selective(client, use_testdb):
    rv = client.get('/datasets?key1=val21')
    assert rv.status_code == 200
    assert len(json.loads(rv.data)['datasets']) == 3

    rv = client.get('/datasets?key1=val21&key2=val23')
    assert rv.status_code == 200
    assert len(json.loads(rv.data)['datasets']) == 1

    rv = client.get('/datasets?key1=[val21]')
    assert rv.status_code == 200
    assert len(json.loads(rv.data)['datasets']) == 3

    rv = client.get('/datasets?key2=[val23,val24]&akey=x')
    assert rv.status_code == 200
    assert len(json.loads(rv.data)['datasets']) == 2


def test_get_datasets_unknown_key(client, use_testdb):
    rv = client.get('/datasets?UNKNOWN=val21')
    assert rv.status_code == 400


def test_get_singleband_greyscale(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png')
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_extra_args(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?foo=bar&baz=quz')
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_cmap(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=jet')
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_preview(client, use_testdb):
    import terracotta
    settings = terracotta.get_settings()

    rv = client.get('/singleband/val11/x/val12/preview.png?colormap=jet')
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def urlsafe_json(payload):
    payload_json = json.dumps(payload)
    return urllib.parse.quote_plus(payload_json, safe=r',.[]{}:"')


def test_get_singleband_explicit_cmap(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    explicit_cmap = {1: (0, 0, 0), 2.0: (255, 255, 255, 20), 3: '#ffffff', 4: 'abcabc'}

    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit'
                    f'&explicit_color_map={urlsafe_json(explicit_cmap)}')
    assert rv.status_code == 200, rv.data.decode('utf-8')

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_explicit_cmap_invalid(client, use_testdb, raster_file_xyz):
    x, y, z = raster_file_xyz
    explicit_cmap = {1: (0, 0, 0), 2: (255, 255, 255), 3: '#ffffff', 4: 'abcabc'}

    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?'
                    f'explicit_color_map={urlsafe_json(explicit_cmap)}')
    assert rv.status_code == 400

    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=jet'
                    f'&explicit_color_map={urlsafe_json(explicit_cmap)}')
    assert rv.status_code == 400

    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit')
    assert rv.status_code == 400

    explicit_cmap[3] = 'omgomg'
    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit'
                    f'&explicit_color_map={urlsafe_json(explicit_cmap)}')
    assert rv.status_code == 400

    explicit_cmap = [(255, 255, 255)]
    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit'
                    f'&explicit_color_map={urlsafe_json(explicit_cmap)}')
    assert rv.status_code == 400

    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit'
                    f'&explicit_color_map=foo')
    assert rv.status_code == 400


def test_get_singleband_stretch(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz

    for stretch_range in ('[0,1]', '[0,null]', '[null, 1]', '[null,null]', 'null'):
        rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?stretch_range={stretch_range}')
        assert rv.status_code == 200

        img = Image.open(BytesIO(rv.data))
        assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_out_of_bounds(client, use_testdb):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = (0, 0, 10)
    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png')
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE
    assert np.all(np.asarray(img) == 0)


def test_get_singleband_unknown_cmap(client, use_testdb, raster_file_xyz):
    x, y, z = raster_file_xyz
    rv = client.get(f'/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=UNKNOWN')
    assert rv.status_code == 400


def test_get_rgb(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f'/rgb/val21/x/{z}/{x}/{y}.png?r=val22&g=val23&b=val24')
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == (*settings.DEFAULT_TILE_SIZE, 3)


def test_get_rgb_preview(client, use_testdb):
    import terracotta
    settings = terracotta.get_settings()

    rv = client.get('/rgb/val21/x/preview.png?r=val22&g=val23&b=val24')
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == (*settings.DEFAULT_TILE_SIZE, 3)


def test_get_rgb_extra_args(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f'/rgb/val21/x/{z}/{x}/{y}.png?r=val22&g=val23&b=val24&foo=bar&baz=quz')
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == (*settings.DEFAULT_TILE_SIZE, 3)


def test_get_rgb_stretch(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz

    for stretch_range in ('[0,10000]', '[0,null]', '[null, 10000]', '[null,null]', 'null'):
        rv = client.get(f'/rgb/val21/x/{z}/{x}/{y}.png?r=val22&g=val23&b=val24&'
                        f'r_range={stretch_range}&b_range={stretch_range}&g_range={stretch_range}')
        assert rv.status_code == 200, rv.data

        img = Image.open(BytesIO(rv.data))
        assert np.asarray(img).shape == (*settings.DEFAULT_TILE_SIZE, 3)


def test_get_compute(client, use_testdb, raster_file_xyz):
    import terracotta
    settings = terracotta.get_settings()

    # default tile size
    x, y, z = raster_file_xyz
    rv = client.get(
        f'/compute/val21/x/{z}/{x}/{y}.png'
        '?expression=v1*v2&v1=val22&v2=val23'
        '&stretch_range=[0,10000]'
    )
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE

    # custom tile size
    rv = client.get(
        f'/compute/val21/x/{z}/{x}/{y}.png'
        '?expression=v1*v2&v1=val22&v2=val23'
        '&stretch_range=[0,10000]'
        '&tile_size=[128,128]'
    )
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == (128, 128)


def test_get_compute_preview(client, use_testdb):
    import terracotta
    settings = terracotta.get_settings()

    rv = client.get(
        '/compute/val21/x/preview.png'
        '?expression=v1*v2&v1=val22&v2=val23'
        '&stretch_range=[0,10000]'
    )
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_compute_invalid(client, use_testdb, raster_file_xyz):
    x, y, z = raster_file_xyz

    # too few keys
    rv = client.get(
        f'/compute/val21/{z}/{x}/{y}.png'
        '?expression=v1*v2&v1=val22&v2=val23'
        '&stretch_range=[0,10000]'
    )
    assert rv.status_code == 400

    # invalid expression
    rv = client.get(
        '/compute/val21/x/preview.png'
        '?expression=__builtins__["dir"](v1)&v1=val22'
        '&stretch_range=[0,10000]'
    )
    assert rv.status_code == 400

    # no stretch range
    rv = client.get(
        f'/compute/val21/x/{z}/{x}/{y}.png'
        '?expression=v1*v2&v1=val22&v2=val23'
    )
    assert rv.status_code == 400

    # no expression
    rv = client.get(
        f'/compute/val21/x/{z}/{x}/{y}.png'
        '?stretch_range=[0,10000)'
    )
    assert rv.status_code == 400

    # missing operand
    rv = client.get(
        f'/compute/val21/x/{z}/{x}/{y}.png'
        '?expression=v1*v2'
        '&stretch_range=[0,10000)'
    )
    assert rv.status_code == 400

    # invalid stretch range (syntax)
    rv = client.get(
        f'/compute/val21/x/{z}/{x}/{y}.png'
        '?expression=v1*v2&v1=val22&v2=val23'
        '&stretch_range=[0,10000)'
    )
    assert rv.status_code == 400

    # invalid stretch range (value)
    rv = client.get(
        f'/compute/val21/x/{z}/{x}/{y}.png'
        '?expression=v1*v2&v1=val22&v2=val23'
        '&stretch_range=[10000,0]'
    )
    assert rv.status_code == 400


def test_get_colormap(client):
    rv = client.get('/colormap?stretch_range=[0,1]&num_values=100')
    assert rv.status_code == 200
    assert len(json.loads(rv.data)['colormap']) == 100


def test_get_colormap_invalid(client):
    rv = client.get('/colormap?stretch_range=[0,1')
    assert rv.status_code == 400


def test_get_colormap_extra_args(client):
    rv = client.get('/colormap?stretch_range=[0,1]&num_values=100&foo=bar&baz=quz')
    assert rv.status_code == 200
    assert len(json.loads(rv.data)['colormap']) == 100


def test_get_spec(client):
    from terracotta import __version__

    rv = client.get('/swagger.json')
    assert rv.status_code == 200
    assert json.loads(rv.data)
    assert __version__ in rv.data.decode('utf-8')

    rv = client.get('/apidoc')
    assert rv.status_code == 200
    assert b'Terracotta' in rv.data



import pytest
from click.testing import CliRunner

import os
import shutil

TEST_CASES = (
    {  # basic
        'filenames': ['foo.tif'],
        'input_pattern': '{name}.tif',
        'expected_keys': ['name'],
        'expected_datasets': [('foo',)]
    },
    {  # two keys
        'filenames': ['S2_20180101_B04.tif'],
        'input_pattern': '{sensor}_{date}_B04.tif',
        'expected_keys': ['sensor', 'date'],
        'expected_datasets': [('S2', '20180101')]
    },
    {  # subfolder
        'filenames': ['S2/20180101_B04.tif'],
        'input_pattern': '{sensor}/{date}_B04.tif',
        'expected_keys': ['sensor', 'date'],
        'expected_datasets': [('S2', '20180101')]
    },
    {  # wildcard magic
        'filenames': ['S2/20180101_B04.tif', 'S2/20180101_B05_median.tif'],
        'input_pattern': '{sensor}/{date}_{band}{}.tif',
        'expected_keys': ['sensor', 'date', 'band'],
        'expected_datasets': [('S2', '20180101', 'B04'), ('S2', '20180101', 'B05')]
    },
    {  # keys occuring more than once
        'filenames': ['S2/20180101/S2_20180101_B04.tif'],
        'input_pattern': '{sensor}/{date}/{sensor}_{date}_{band}.tif',
        'expected_keys': ['sensor', 'date', 'band'],
        'expected_datasets': [('S2', '20180101', 'B04')]
    },
    {  # {} in filename
        'filenames': ['bar.tif', '{foo}.tif'],
        'input_pattern': '{{{name}}}.tif',
        'expected_keys': ['name'],
        'expected_datasets': [('foo',)]
    },
    {  # case-sensitivity
        'filenames': ['bAr.tif', 'FOO.tif'],
        'input_pattern': '{nAmE}.tif',
        'expected_keys': ['nAmE'],
        'expected_datasets': [('bAr',), ('FOO',)]
    },
    {  # unicode path with space
        'filenames': ['süßer ordner/foo.tif'],
        'input_pattern': '{}/{name}.tif',
        'expected_keys': ['name'],
        'expected_datasets': [('foo',)]
    },
    {  # unicode key
        'filenames': ['günther.tif'],
        'input_pattern': '{bärbel}.tif',
        'expected_keys': ['bärbel'],
        'expected_datasets': [('günther',)]
    },
)

INVALID_TEST_CASES = (
    {  # no matching files
        'filenames': [],
        'input_pattern': '{key}.tif',
        'error_contains': 'matches no files'
    },
    {  # duplicate keys in different folders
        'filenames': ['dir1/foo.tif', 'dir2/foo.tif'],
        'input_pattern': '{}/{name}.tif',
        'error_contains': 'duplicate keys'
    },
    {  # duplicate keys through wildcard
        'filenames': ['S2_B04.tif', 'S2_20180101_B04.tif'],
        'input_pattern': '{sensor}_{}.tif',
        'error_contains': 'duplicate keys'
    },
    {  # no groups in pattern
        'filenames': [],
        'input_pattern': 'notafile.tif',
        'error_contains': 'at least one placeholder'
    },
    {  # only wildcards in pattern
        'filenames': [],
        'input_pattern': '{}.tif',
        'error_contains': 'at least one placeholder'
    },
    {  # stray {
        'filenames': [],
        'input_pattern': 'notafile{.tif',
        'error_contains': 'invalid pattern'
    },
    {  # invalid placeholder name
        'filenames': ['foo.tif'],
        'input_pattern': '{(foo)}.tif',
        'error_contains': 'must be alphanumeric'
    }
)


@pytest.fixture()
def tmpworkdir(tmpdir):
    orig_dir = os.getcwd()
    try:
        os.chdir(tmpdir)
        yield tmpdir
    finally:
        os.chdir(orig_dir)


def test_ingest(raster_file, tmpdir):
    from terracotta.scripts import cli

    outfile = tmpdir / 'out.sqlite'
    input_pattern = str(raster_file.dirpath('{name}.tif'))

    runner = CliRunner()
    result = runner.invoke(cli.cli, ['ingest', input_pattern, '-o', str(outfile)])
    assert result.exit_code == 0
    assert outfile.check()

    from terracotta import get_driver
    driver = get_driver(str(outfile), provider='sqlite')
    assert driver.key_names == ('name',)
    assert driver.get_datasets() == {('img',): str(raster_file)}


def test_ingest_append(raster_file, tmpworkdir):
    from terracotta.scripts import cli

    for infile in ('dir1/img1.tif', 'dir2/img2.tif'):
        temp_infile = tmpworkdir / infile
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / 'out.sqlite'

    runner = CliRunner()
    result = runner.invoke(cli.cli, ['ingest', 'dir1/{name}.tif', '-o', str(outfile)])
    assert result.exit_code == 0
    assert outfile.check()

    result = runner.invoke(cli.cli, ['ingest', 'dir2/{name}.tif', '-o', str(outfile)])
    assert result.exit_code == 0
    assert outfile.check()

    from terracotta import get_driver
    driver = get_driver(str(outfile), provider='sqlite')
    assert driver.key_names == ('name',)
    assert all((ds,) in driver.get_datasets() for ds in ('img1', 'img2'))


@pytest.mark.parametrize('skip_existing', [True, False])
def test_reingest(skip_existing, raster_file, tmpworkdir):
    from terracotta.scripts import cli
    from terracotta import get_driver

    same_name = 'myimage'
    infiles = [
        tmpworkdir / p
        for p in [f'dir1/{same_name}.tif', f'dir2/{same_name}.tif', f'dir3/{same_name}.tif']
    ]
    for temp_infile in infiles:
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / 'out.sqlite'

    def _assert_datasets_equal(datasets):
        assert outfile.check()
        driver = get_driver(str(outfile), provider='sqlite')
        existing_datasets = driver.get_datasets()
        assert len(existing_datasets) == 1
        assert existing_datasets == datasets

    runner = CliRunner()
    args = ['ingest', 'dir1/{name}.tif', '-o', str(outfile)]
    result = runner.invoke(cli.cli, args)
    assert result.exit_code == 0
    _assert_datasets_equal({(same_name,): str(infiles[0])})

    args = ['ingest', 'dir2/{name}.tif', '-o', str(outfile)]
    if skip_existing:
        args.append("--skip-existing")
    result = runner.invoke(cli.cli, args)
    assert result.exit_code == 0
    if skip_existing:
        _assert_datasets_equal({(same_name,): str(infiles[0])})
    else:
        _assert_datasets_equal({(same_name,): str(infiles[1])})

    args = ['ingest', 'dir3/{name}.tif', '-o', str(outfile)]
    if skip_existing:
        args.append("--skip-existing")
    result = runner.invoke(cli.cli, args)
    assert result.exit_code == 0
    if skip_existing:
        _assert_datasets_equal({(same_name,): str(infiles[0])})
    else:
        _assert_datasets_equal({(same_name,): str(infiles[2])})


def test_ingest_append_invalid(raster_file, tmpworkdir):
    from terracotta.scripts import cli

    for infile in ('dir1/img1.tif', 'dir2/img2.tif'):
        temp_infile = tmpworkdir / infile
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / 'out.sqlite'

    runner = CliRunner()
    result = runner.invoke(cli.cli, ['ingest', 'dir1/{name}.tif', '-o', str(outfile)])
    assert result.exit_code == 0
    assert outfile.check()

    result = runner.invoke(cli.cli, ['ingest', '{dir}/{name}.tif', '-o', str(outfile)])
    assert result.exit_code != 0
    assert 'incompatible key names' in result.output


@pytest.mark.parametrize('case', TEST_CASES)
@pytest.mark.parametrize('abspath', [True, False])
def test_ingest_pattern(case, abspath, raster_file, tmpworkdir):
    from terracotta.scripts import cli

    for infile in case['filenames']:
        temp_infile = tmpworkdir / infile
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / 'out.sqlite'

    if abspath:
        input_pattern = os.path.abspath(tmpworkdir / case['input_pattern'])
    else:
        input_pattern = case['input_pattern']

    runner = CliRunner()
    result = runner.invoke(cli.cli, ['ingest', input_pattern, '-o', str(outfile)])
    assert result.exit_code == 0, result.output
    assert outfile.check()

    from terracotta import get_driver
    driver = get_driver(str(outfile), provider='sqlite')
    assert driver.key_names == tuple(case['expected_keys'])
    assert all(ds in driver.get_datasets() for ds in case['expected_datasets'])


@pytest.mark.parametrize('case', INVALID_TEST_CASES)
def test_ingest_invalid_pattern(case, raster_file, tmpworkdir):
    from terracotta.scripts import cli

    for infile in case['filenames']:
        temp_infile = tmpworkdir / infile
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / 'out.sqlite'
    input_pattern = case['input_pattern']

    runner = CliRunner()
    result = runner.invoke(cli.cli, ['ingest', input_pattern, '-o', str(outfile)])
    assert result.exit_code != 0
    assert case['error_contains'].lower() in result.output.lower()


def test_ingest_rgb_key(raster_file, tmpdir):
    from terracotta.scripts import cli

    outfile = tmpdir / 'out.sqlite'
    input_pattern = str(raster_file.dirpath('{rgb}m{foo}.tif'))

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ['ingest', input_pattern, '-o', str(outfile), '--rgb-key', 'rgb']
    )
    assert result.exit_code == 0
    assert outfile.check()

    from terracotta import get_driver
    driver = get_driver(str(outfile), provider='sqlite')
    assert driver.key_names == ('foo', 'rgb')
    assert driver.get_datasets() == {('g', 'i'): str(raster_file)}


def test_ingest_invalid_rgb_key(raster_file, tmpdir):
    from terracotta.scripts import cli

    outfile = tmpdir / 'out.sqlite'
    input_pattern = str(raster_file.dirpath('{rgb}m{foo}.tif'))

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ['ingest', input_pattern, '-o', str(outfile), '--rgb-key', 'bar']
    )
    assert result.exit_code != 0
    assert not outfile.check()



import os
import warnings
import traceback

import rasterio
import numpy as np

from click.testing import CliRunner

import pytest


def format_exception(result):
    return ''.join(traceback.format_exception(*result.exc_info))


@pytest.fixture()
def tiny_raster_file(unoptimized_raster_file, tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp('tiny-raster')
    outfile = tmpdir / 'tiny.tif'
    with rasterio.open(str(unoptimized_raster_file)) as src:
        profile = src.profile.copy()
        profile.update(
            width=100,
            height=100,
            blockxsize=256,
            blockysize=256
        )

        with rasterio.open(str(outfile), 'w', **profile) as dst:
            dst.write(src.read()[:100, :100])

    yield outfile


@pytest.mark.parametrize('in_memory', [True, None, False])
@pytest.mark.parametrize('reproject', [True, False])
@pytest.mark.parametrize('compression', ['auto', 'lzw', 'none'])
@pytest.mark.parametrize('nproc', [None, 1, 2, -1])
def test_optimize_rasters(unoptimized_raster_file, tmpdir, in_memory,
                          reproject, compression, nproc):
    from terracotta.cog import validate
    from terracotta.scripts import cli

    input_pattern = str(unoptimized_raster_file.dirpath('*.tif'))
    outfile = tmpdir / unoptimized_raster_file.basename

    runner = CliRunner()

    flags = ['--compression', compression, '-q']

    if in_memory is not None:
        flags.append('--in-memory' if in_memory else '--no-in-memory')

    if reproject:
        flags.append('--reproject')

    if nproc is not None:
        flags.append(f'--nproc={nproc}')

    result = runner.invoke(cli.cli, ['optimize-rasters', input_pattern, '-o', str(tmpdir), *flags])

    assert result.exit_code == 0, format_exception(result)
    assert outfile.check()

    # validate files
    assert not validate(str(unoptimized_raster_file))
    assert validate(str(outfile))

    if reproject:
        return

    # check for data integrity
    with rasterio.open(str(unoptimized_raster_file)) as src1, rasterio.open(str(outfile)) as src2:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'invalid value encountered.*')
            np.testing.assert_array_equal(src1.read(), src2.read())


def test_optimize_rasters_small(tiny_raster_file, tmpdir):
    from terracotta.cog import validate
    from terracotta.scripts import cli

    input_pattern = str(tiny_raster_file)
    outfile = tmpdir / tiny_raster_file.basename

    runner = CliRunner()
    result = runner.invoke(cli.cli, ['optimize-rasters', input_pattern, '-o', str(tmpdir)])

    assert result.exit_code == 0, format_exception(result)
    assert outfile.check()

    # validate files
    # (small rasters don't need overviews, so input file is valid, too)
    assert validate(str(tiny_raster_file))
    assert validate(str(outfile))

    # check for data integrity
    with rasterio.open(str(tiny_raster_file)) as src1, rasterio.open(str(outfile)) as src2:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'invalid value encountered.*')
            np.testing.assert_array_equal(src1.read(), src2.read())


def test_optimize_rasters_nofiles(tmpdir):
    from terracotta.scripts import cli

    input_pattern = str(tmpdir.dirpath('*.tif'))
    runner = CliRunner()
    result = runner.invoke(cli.cli, ['optimize-rasters', input_pattern, '-o', str(tmpdir), '-q'])

    assert result.exit_code == 0
    assert 'No files given' in result.output


def test_optimize_rasters_invalid(tmpdir):
    from terracotta.scripts import cli

    runner = CliRunner()
    result = runner.invoke(cli.cli, ['optimize-rasters', str(tmpdir), '-o', str(tmpdir), '-q'])

    assert result.exit_code != 0
    assert 'not a file' in result.output

    result = runner.invoke(cli.cli, ['optimize-rasters', str(tmpdir), '-o', str(tmpdir),
                                     '--overwrite', '--skip-existing'])
    assert result.exit_code != 0
    assert 'mutually exclusive' in result.output


def test_optimize_rasters_multiband(tmpdir, unoptimized_raster_file):
    from terracotta.scripts import cli
    import rasterio

    with rasterio.open(str(unoptimized_raster_file)) as src:
        profile = src.profile.copy()
        data = src.read(1)

    profile['count'] = 3

    multiband_file = tmpdir.join(unoptimized_raster_file.basename)
    with rasterio.open(str(multiband_file), 'w', **profile) as dst:
        dst.write(data, 1)
        dst.write(data, 2)
        dst.write(data, 3)

    input_pattern = str(multiband_file.dirpath('*.tif'))
    outfile = tmpdir / 'co' / unoptimized_raster_file.basename

    runner = CliRunner()
    result = runner.invoke(
        cli.cli,
        ['optimize-rasters', input_pattern, '-o', str(tmpdir / 'co')]
    )

    assert result.exit_code == 0
    assert 'has more than one band' in result.output

    with rasterio.open(str(unoptimized_raster_file)) as src1, rasterio.open(str(outfile)) as src2:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'invalid value encountered.*')
            np.testing.assert_array_equal(src1.read(), src2.read())


@pytest.mark.parametrize('extra_flag', ['skip-existing', 'overwrite', None])
def test_reoptimize(tmpdir, unoptimized_raster_file, extra_flag):
    from terracotta.scripts import cli

    infile = str(unoptimized_raster_file.dirpath('*.tif'))
    outfile = tmpdir / 'out.tif'

    # first time
    runner = CliRunner()
    args = ['optimize-rasters', infile, '-o', str(outfile)]
    result = runner.invoke(cli.cli, args)
    assert result.exit_code == 0
    ctime = os.path.getmtime(outfile)

    # second time
    args = ['optimize-rasters', infile, '-o', str(outfile)]
    if extra_flag:
        args.append(f'--{extra_flag}')

    result = runner.invoke(cli.cli, args)

    if extra_flag == 'skip-existing':
        assert result.exit_code == 0
        assert os.path.getmtime(outfile) == ctime
    elif extra_flag == 'overwrite':
        assert result.exit_code == 0
        assert os.path.getmtime(outfile) != ctime
    else:
        assert result.exit_code == 2


def _throw(*args):
    raise RuntimeError('A mock error is raised')


def test_exception_in_subprocess(unoptimized_raster_file, tmpdir, monkeypatch):
    from terracotta.scripts import cli

    monkeypatch.setattr(
        'terracotta.scripts.optimize_rasters._optimize_single_raster',
        _throw
    )

    args = [
        'optimize-rasters', str(unoptimized_raster_file), '-o',
        str(tmpdir / 'foo.tif'), '--nproc', 2
    ]

    runner = CliRunner()
    result = runner.invoke(cli.cli, args)

    assert result.exit_code != 0
    assert 'Error while optimizing file' in str(result.exception)



import pytest

import time

from moto import mock_xray_client, XRaySegment


def test_xray_tracing(caplog):
    @mock_xray_client
    def run_test():
        from terracotta import update_settings
        import terracotta.profile

        update_settings(XRAY_PROFILE=True)

        @terracotta.profile.trace('dummy')
        def func_to_trace():
            time.sleep(0.1)

        with XRaySegment():
            func_to_trace()

        with XRaySegment():
            with terracotta.profile.trace('dummy2'):
                time.sleep(0.1)

        for record in caplog.records:
            assert record.levelname != 'ERROR'

        # sanity check, recording without starting a segment should fail
        func_to_trace()
        assert any('cannot find the current segment' in rec.message for rec in caplog.records)

    run_test()


def test_xray_exception(caplog):
    @mock_xray_client
    def run_test():
        from terracotta import update_settings
        import terracotta.profile

        update_settings(XRAY_PROFILE=True)

        with XRaySegment():
            with pytest.raises(NotImplementedError):
                with terracotta.profile.trace('dummy') as subsegment:
                    raise NotImplementedError('foo')

        assert len(subsegment.cause['exceptions']) == 1
        assert subsegment.cause['exceptions'][0].message == 'foo'

    run_test()



import pytest

import numpy as np


@pytest.fixture(autouse=True)
def reload_cmap():
    import importlib
    import terracotta.cmaps.get_cmaps
    try:
        yield
    finally:
        importlib.reload(terracotta.cmaps.get_cmaps)


def test_get_cmap():
    from terracotta.cmaps.get_cmaps import get_cmap, AVAILABLE_CMAPS
    for name in AVAILABLE_CMAPS:
        cmap = get_cmap(name)
        assert cmap.shape == (255, 4)
        assert cmap.dtype == np.uint8


def test_get_cmap_filesystem(monkeypatch):
    import pkg_resources
    import importlib

    import terracotta.cmaps.get_cmaps

    def throw_error(*args, **kwargs):
        raise pkg_resources.DistributionNotFound('monkeypatched')

    with monkeypatch.context() as m:
        m.setattr(pkg_resources.Requirement, 'parse', throw_error)

        with pytest.raises(pkg_resources.DistributionNotFound):
            pkg_resources.Requirement.parse('terracotta')

        importlib.reload(terracotta.cmaps.get_cmaps)

        cmap = terracotta.cmaps.get_cmaps.get_cmap('jet')
        assert cmap.shape == (255, 4)
        assert cmap.dtype == np.uint8


def test_extra_cmap(monkeypatch, tmpdir):
    import importlib

    import terracotta.cmaps.get_cmaps

    custom_cmap_data = np.tile(
        np.arange(255, dtype='uint8'),
        (4, 1)
    ).T
    np.save(str(tmpdir / f'foo{terracotta.cmaps.get_cmaps.SUFFIX}'), custom_cmap_data)
    np.save(str(tmpdir / 'bar.npy'), custom_cmap_data)

    with monkeypatch.context() as m:
        m.setenv('TC_EXTRA_CMAP_FOLDER', str(tmpdir))
        importlib.reload(terracotta.cmaps.get_cmaps)

        assert 'foo' in terracotta.cmaps.get_cmaps.AVAILABLE_CMAPS
        assert 'bar' not in terracotta.cmaps.get_cmaps.AVAILABLE_CMAPS

        np.testing.assert_equal(
            custom_cmap_data,
            terracotta.cmaps.get_cmaps.get_cmap('foo')
        )


def test_extra_cmap_invalid_shape(monkeypatch, tmpdir):
    import importlib
    import terracotta.cmaps.get_cmaps

    broken_cmap_data = np.tile(
        np.arange(666, dtype='uint8'),
        (4, 1)
    ).T
    np.save(str(tmpdir / f'foo{terracotta.cmaps.get_cmaps.SUFFIX}'), broken_cmap_data)

    with monkeypatch.context() as m:
        m.setenv('TC_EXTRA_CMAP_FOLDER', str(tmpdir))

        with pytest.raises(ValueError) as raised_exc:
            importlib.reload(terracotta.cmaps.get_cmaps)

        assert 'foo' in str(raised_exc.value)
        assert '666' in str(raised_exc.value)


def test_extra_cmap_invalid_folder(monkeypatch):
    import importlib
    import terracotta.cmaps.get_cmaps

    with monkeypatch.context() as m:
        m.setenv('TC_EXTRA_CMAP_FOLDER', 'bar')

        with pytest.raises(IOError) as raised_exc:
            importlib.reload(terracotta.cmaps.get_cmaps)

        assert 'bar' in str(raised_exc.value)


def test_extra_cmap_invalid_dtype(monkeypatch, tmpdir):
    import importlib
    import terracotta.cmaps.get_cmaps

    broken_cmap_data = np.tile(
        np.arange(255, dtype='float'),
        (4, 1)
    ).T
    np.save(str(tmpdir / f'foo{terracotta.cmaps.get_cmaps.SUFFIX}'), broken_cmap_data)

    with monkeypatch.context() as m:
        m.setenv('TC_EXTRA_CMAP_FOLDER', str(tmpdir))

        with pytest.raises(ValueError) as raised_exc:
            importlib.reload(terracotta.cmaps.get_cmaps)

        assert 'foo' in str(raised_exc.value)
        assert 'float' in str(raised_exc.value)



"""Benchmarks for all performance-critical use cases

Run separately via `pytest tests/benchmarks.py`.
"""

import pytest
from click.testing import CliRunner

ZOOM_XYZ = {
    'birds-eye': 8,
    'balanced': 12,
    'subpixel': 21,
    'preview': None
}


@pytest.fixture(scope='session')
def benchmark_database(big_raster_file_nodata, big_raster_file_mask, tmpdir_factory):
    from terracotta import get_driver, update_settings

    keys = ['type', 'band']

    update_settings(RASTER_CACHE_SIZE=0)

    dbpath = tmpdir_factory.mktemp('db').join('db-readonly.sqlite')
    driver = get_driver(dbpath, provider='sqlite')
    driver.create(keys)

    mtd = driver.compute_metadata(str(big_raster_file_nodata))

    with driver.connect():
        driver.insert(['nodata', '1'], str(big_raster_file_nodata), metadata=mtd)
        driver.insert(['nodata', '2'], str(big_raster_file_nodata), metadata=mtd)
        driver.insert(['nodata', '3'], str(big_raster_file_nodata), metadata=mtd)
        driver.insert(['mask', '1'], str(big_raster_file_mask), metadata=mtd)

    return dbpath


def get_xyz(raster_file, zoom):
    import rasterio
    import rasterio.warp
    import mercantile

    with rasterio.open(str(raster_file)) as src:
        raster_bounds = rasterio.warp.transform_bounds(src.crs, 'epsg:4326', *src.bounds)
    raster_center_x = (raster_bounds[0] + raster_bounds[2]) / 2
    raster_center_y = (raster_bounds[1] + raster_bounds[3]) / 2

    tile = mercantile.tile(raster_center_x, raster_center_y, zoom)
    return (tile.x, tile.y, zoom)


@pytest.mark.parametrize('resampling', ['nearest', 'linear', 'cubic', 'average'])
@pytest.mark.parametrize('zoom', ZOOM_XYZ.keys())
def test_bench_rgb(benchmark, zoom, resampling, big_raster_file_nodata, benchmark_database):
    from terracotta.server import create_app
    from terracotta import update_settings

    update_settings(
        DRIVER_PATH=str(benchmark_database),
        RESAMPLING_METHOD=resampling,
        REPROJECTION_METHOD=resampling
    )

    zoom_level = ZOOM_XYZ[zoom]

    flask_app = create_app()
    with flask_app.test_client() as client:
        if zoom_level is not None:
            x, y, z = get_xyz(big_raster_file_nodata, zoom_level)
            rv = benchmark(client.get, f'/rgb/nodata/{z}/{x}/{y}.png?r=1&g=2&b=3')
        else:
            rv = benchmark(client.get, '/rgb/nodata/preview.png?r=1&g=2&b=3')

    assert rv.status_code == 200


def test_bench_rgb_out_of_bounds(benchmark, big_raster_file_nodata, benchmark_database):
    from terracotta.server import create_app
    from terracotta import update_settings

    update_settings(DRIVER_PATH=str(benchmark_database))

    x, y, z = 0, 0, 20

    flask_app = create_app()
    with flask_app.test_client() as client:
        rv = benchmark(client.get, f'/rgb/nodata/{z}/{x}/{y}.png?r=1&g=2&b=3')

    assert rv.status_code == 200


@pytest.mark.parametrize('resampling', ['nearest', 'linear', 'cubic', 'average'])
@pytest.mark.parametrize('zoom', ZOOM_XYZ.keys())
def test_bench_singleband(benchmark, zoom, resampling, big_raster_file_nodata, benchmark_database):
    from terracotta.server import create_app
    from terracotta import update_settings, get_driver

    update_settings(
        DRIVER_PATH=str(benchmark_database),
        RESAMPLING_METHOD=resampling,
        REPROJECTION_METHOD=resampling
    )

    zoom_level = ZOOM_XYZ[zoom]

    flask_app = create_app()
    with flask_app.test_client() as client:
        if zoom_level is not None:
            x, y, z = get_xyz(big_raster_file_nodata, zoom_level)
            rv = benchmark(client.get, f'/singleband/nodata/1/{z}/{x}/{y}.png')
        else:
            rv = benchmark(client.get, '/singleband/nodata/1/preview.png')

    assert rv.status_code == 200
    assert not len(get_driver(str(benchmark_database))._raster_cache)


def test_bench_singleband_out_of_bounds(benchmark, benchmark_database):
    from terracotta.server import create_app
    from terracotta import update_settings

    update_settings(DRIVER_PATH=str(benchmark_database))

    x, y, z = 0, 0, 20

    flask_app = create_app()
    with flask_app.test_client() as client:
        rv = benchmark(client.get, f'/singleband/nodata/1/{z}/{x}/{y}.png')

    assert rv.status_code == 200


@pytest.mark.parametrize('chunks', [False, True])
@pytest.mark.parametrize('raster_type', ['nodata', 'masked'])
def test_bench_compute_metadata(benchmark, big_raster_file_nodata, big_raster_file_mask,
                                chunks, raster_type):
    from terracotta.drivers.raster_base import RasterDriver
    if raster_type == 'nodata':
        raster_file = big_raster_file_nodata
    elif raster_type == 'masked':
        raster_file = big_raster_file_mask
    benchmark(RasterDriver.compute_metadata, str(raster_file), use_chunks=chunks)


@pytest.mark.parametrize('in_memory', [False, True])
def test_bench_optimize_rasters(benchmark, unoptimized_raster_file, tmpdir, in_memory):
    from terracotta.scripts import cli

    input_pattern = str(unoptimized_raster_file.dirpath('*.tif'))
    in_memory_flag = '--in-memory' if in_memory else '--no-in-memory'

    runner = CliRunner()
    result = benchmark(
        runner.invoke, cli.cli,
        ['optimize-rasters', input_pattern, '-o', str(tmpdir), in_memory_flag, '--overwrite']
    )

    assert result.exit_code == 0



from setuptools import setup, find_packages
from os import path
import re

here = path.abspath(path.dirname(__file__))

# get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# resolve relative links to figures
long_description = re.sub(
    r'!\[(.*?)\]\((docs/_figures/.*?)\)',
    r'![\1](https://github.com/DHI-GRAS/terracotta/raw/main/\2)',
    long_description
)

numpy_version = '>=1.15,!=1.17.0'

setup(
    # metadata
    name='terracotta',
    description='A modern XYZ tile server written in Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/DHI-GRAS/terracotta',
    author='Jonas Sølvsteen',
    author_email='josl@dhigroup.com',
    keywords='xyz tileserver rasterio cloud-optimized-geotiff serverless',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Flask',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Multimedia :: Graphics :: Viewers',
        'Topic :: Scientific/Engineering :: GIS'
    ],
    # module
    packages=find_packages(exclude=['docs', 'tests']),
    python_requires='>=3.6',
    use_scm_version={
        'write_to': 'terracotta/_version.py'
    },
    # dependencies
    setup_requires=[
        'setuptools_scm',
        'setuptools_scm_git_archive',
        'numpy%s' % numpy_version
    ],
    install_requires=[
        'apispec>=1.0',
        'apispec-webframeworks',
        'cachetools>=3.1.0',
        'click',
        'click-spinner',
        'flask',
        'flask_cors',
        'marshmallow>=3.0.0',
        'mercantile',
        'numpy%s' % numpy_version,
        'pillow',
        'pyyaml>=3.10',  # downstream dependency of apispec
        'shapely',
        'rasterio>=1.0',
        'shapely',
        'toml',
        'tqdm'
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-cov',
            'pytest-mypy',
            'pytest-flake8',
            'pytest-benchmark',
            'attrs>=17.4.0',
            'codecov',
            'colorlog',
            'crick',
            'flake8',
            'matplotlib',
            'moto',
            'aws-xray-sdk',
            'pymysql>=1.0.0'
        ],
        'docs': [
            'sphinx',
            'sphinx_autodoc_typehints',
            'sphinx-click',
            'pymysql>=1.0.0'
        ],
        'recommended': [
            'colorlog',
            'crick',
            'pymysql>=1.0.0'
        ]
    },
    # CLI
    entry_points='''
        [console_scripts]
        terracotta=terracotta.scripts.cli:entrypoint
    ''',
    # package data
    include_package_data=True,
    package_data={
        'terracotta': [
            'cmaps/data/*_rgba.npy',  # colormaps
            'templates/*.html', 'static/*'  # preview app
        ]
    },
)



from PIL import Image
import numpy as np

import pytest


def test_rgb_handler(use_testdb, raster_file, raster_file_xyz):
    import terracotta
    from terracotta.handlers import rgb
    raw_img = rgb.rgb(['val21', 'x'], ['val22', 'val23', 'val24'], raster_file_xyz)
    img_data = np.asarray(Image.open(raw_img))
    assert img_data.shape == (*terracotta.get_settings().DEFAULT_TILE_SIZE, 3)


def test_rgb_tile_size(use_testdb, raster_file, raster_file_xyz):
    from terracotta.handlers import rgb
    raw_img = rgb.rgb(['val21', 'x'], ['val22', 'val23', 'val24'], raster_file_xyz,
                      tile_size=(100, 100))
    img_data = np.asarray(Image.open(raw_img))
    assert img_data.shape == (100, 100, 3)


def test_rgb_invalid_keys(use_testdb, raster_file_xyz):
    from terracotta import exceptions
    from terracotta.handlers import rgb

    with pytest.raises(exceptions.InvalidArgumentsError):
        rgb.rgb(['val21', 'x', 'y', 'z'], ['val22', 'val23', 'val24'], raster_file_xyz)


def test_rgb_invalid_args(use_testdb, raster_file_xyz):
    from terracotta import exceptions
    from terracotta.handlers import rgb

    with pytest.raises(exceptions.InvalidArgumentsError):
        rgb.rgb(['val21', 'x'], ['val22', 'val23', 'val24'], raster_file_xyz, stretch_ranges=[])


def test_rgb_invalid_rgb_values(use_testdb, raster_file_xyz):
    from terracotta import exceptions
    from terracotta.handlers import rgb

    with pytest.raises(exceptions.InvalidArgumentsError):
        rgb.rgb(['val21', 'x'], ['val22', 'val23'], raster_file_xyz)


def test_rgb_out_of_bounds(use_testdb, raster_file):
    import terracotta
    from terracotta.handlers import rgb

    with pytest.raises(terracotta.exceptions.TileOutOfBoundsError) as excinfo:
        rgb.rgb(['val21', 'x'], ['val22', 'val23', 'val24'], (10, 0, 0))
        assert 'data covers less than' not in str(excinfo.value)


def test_rgb_lowzoom(use_testdb, raster_file, raster_file_xyz_lowzoom):
    import terracotta
    from terracotta.handlers import rgb

    with pytest.raises(terracotta.exceptions.TileOutOfBoundsError) as excinfo:
        rgb.rgb(['val21', 'x'], ['val22', 'val23', 'val24'], raster_file_xyz_lowzoom)
        assert 'data covers less than' in str(excinfo.value)


@pytest.mark.parametrize('stretch_range', [[0, 20000], [10000, 20000], [-50000, 50000], [100, 100]])
def test_rgb_stretch(stretch_range, use_testdb, testdb, raster_file_xyz):
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import rgb

    ds_keys = ['val21', 'x', 'val22']

    raw_img = rgb.rgb(ds_keys[:2], ['val22', 'val23', 'val24'], raster_file_xyz,
                      stretch_ranges=[stretch_range] * 3)
    img_data = np.asarray(Image.open(raw_img))[..., 0]

    # get unstretched data to compare to
    driver = terracotta.get_driver(testdb)

    with driver.connect():
        tile_data = get_tile_data(driver, ds_keys, tile_xyz=raster_file_xyz,
                                  tile_size=img_data.shape)

    # filter transparent values
    valid_mask = ~tile_data.mask
    assert np.all(img_data[~valid_mask] == 0)

    valid_img = img_data[valid_mask]
    valid_data = tile_data.compressed()

    assert np.all(valid_img[valid_data < stretch_range[0]] == 1)
    stretch_range_mask = (valid_data > stretch_range[0]) & (valid_data < stretch_range[1])
    assert np.all(valid_img[stretch_range_mask] >= 1)
    assert np.all(valid_img[stretch_range_mask] <= 255)
    assert np.all(valid_img[valid_data > stretch_range[1]] == 255)


def test_rgb_invalid_stretch(use_testdb, raster_file_xyz):
    from terracotta import exceptions
    from terracotta.handlers import rgb

    stretch_range = [100, 0]
    ds_keys = ['val21', 'x', 'val22']

    with pytest.raises(exceptions.InvalidArgumentsError):
        rgb.rgb(ds_keys[:2], ['val22', 'val23', 'val24'], raster_file_xyz,
                stretch_ranges=[stretch_range] * 3)


def test_rgb_preview(use_testdb):
    import terracotta
    from terracotta.handlers import rgb
    raw_img = rgb.rgb(['val21', 'x'], ['val22', 'val23', 'val24'])
    img_data = np.asarray(Image.open(raw_img))
    assert img_data.shape == (*terracotta.get_settings().DEFAULT_TILE_SIZE, 3)



import pytest

import affine
import rasterio
from rasterio.io import MemoryFile
from rasterio.shutil import copy
from rasterio.enums import Resampling

import numpy as np

BASE_PROFILE = {
    'driver': 'GTiff',
    'dtype': 'uint16',
    'nodata': 0,
    'count': 1,
    'crs': {'init': 'epsg:32637'},
    'transform': affine.Affine(
        2.0, 0.0, 694920.0,
        0.0, -2.0, 2055666.0
    )
}


def test_validate_optimized(tmpdir):
    from terracotta import cog

    outfile = str(tmpdir / 'raster.tif')
    raster_data = 1000 * np.random.rand(512, 512).astype(np.uint16)

    profile = BASE_PROFILE.copy()
    profile.update(
        height=raster_data.shape[0],
        width=raster_data.shape[1],
        tiled=True,
        blockxsize=256,
        blockysize=256
    )

    with MemoryFile() as mf, mf.open(**profile) as dst:
        dst.write(raster_data, 1)

        overviews = [2 ** j for j in range(1, 4)]
        dst.build_overviews(overviews, Resampling.nearest)

        copy(dst, outfile, copy_src_overviews=True, **profile)

    assert cog.validate(outfile)


def test_validate_optimized_small(tmpdir):
    from terracotta import cog

    outfile = str(tmpdir / 'raster.tif')
    raster_data = 1000 * np.random.rand(128, 128).astype(np.uint16)

    profile = BASE_PROFILE.copy()
    profile.update(
        height=raster_data.shape[0],
        width=raster_data.shape[1]
    )

    with rasterio.open(outfile, 'w', **profile) as dst:
        dst.write(raster_data, 1)

    assert cog.validate(outfile)


def test_validate_unoptimized(tmpdir):
    from terracotta import cog

    outfile = str(tmpdir / 'raster.tif')
    raster_data = 1000 * np.random.rand(512, 512).astype(np.uint16)

    profile = BASE_PROFILE.copy()
    profile.update(
        height=raster_data.shape[0],
        width=raster_data.shape[1]
    )

    with rasterio.open(outfile, 'w', **profile) as dst:
        dst.write(raster_data, 1)

    assert not cog.validate(outfile)


def test_validate_no_overviews(tmpdir):
    from terracotta import cog

    outfile = str(tmpdir / 'raster.tif')
    raster_data = 1000 * np.random.rand(512, 512).astype(np.uint16)

    profile = BASE_PROFILE.copy()
    profile.update(
        height=raster_data.shape[0],
        width=raster_data.shape[1],
        tiled=True,
        blockxsize=256,
        blockysize=256
    )

    with rasterio.open(outfile, 'w', **profile) as dst:
        dst.write(raster_data, 1)

    assert not cog.validate(outfile)


def test_validate_not_tiled(tmpdir):
    from terracotta import cog

    outfile = str(tmpdir / 'raster.tif')
    raster_data = 1000 * np.random.rand(512, 512).astype(np.uint16)

    profile = BASE_PROFILE.copy()
    profile.update(
        height=raster_data.shape[0],
        width=raster_data.shape[1]
    )

    with rasterio.open(outfile, 'w', **profile) as dst:
        dst.write(raster_data, 1)

        overviews = [2 ** j for j in range(1, 4)]
        dst.build_overviews(overviews, Resampling.nearest)

    assert not cog.validate(outfile)


def test_validate_wrong_offset(tmpdir):
    from terracotta import cog

    outfile = str(tmpdir / 'raster.tif')
    raster_data = 1000 * np.random.rand(512, 512).astype(np.uint16)

    profile = BASE_PROFILE.copy()
    profile.update(
        height=raster_data.shape[0],
        width=raster_data.shape[1],
        tiled=True,
        blockxsize=256,
        blockysize=256
    )

    with rasterio.open(outfile, 'w', **profile) as dst:
        dst.write(raster_data, 1)

        overviews = [2 ** j for j in range(1, 4)]
        dst.build_overviews(overviews, Resampling.nearest)

    assert not cog.validate(outfile)


def test_validate_external_overview(tmpdir):
    import os
    from terracotta import cog

    outfile = str(tmpdir / 'raster.tif')
    raster_data = 1000 * np.random.rand(512, 512).astype(np.uint16)

    profile = BASE_PROFILE.copy()
    profile.update(
        height=raster_data.shape[0],
        width=raster_data.shape[1],
        tiled=True,
        blockxsize=256,
        blockysize=256
    )

    with rasterio.Env(TIFF_USE_OVR=True):
        with rasterio.open(outfile, 'w', **profile) as dst:
            dst.write(raster_data, 1)

            overviews = [2 ** j for j in range(1, 4)]
            dst.build_overviews(overviews, Resampling.nearest)

        assert os.path.isfile(f'{outfile}.ovr')

    assert not cog.validate(outfile)


@pytest.mark.filterwarnings("ignore:Dataset has no geotransform set")
def test_validate_not_gtiff(tmpdir):
    from terracotta import cog

    outfile = str(tmpdir / 'raster.png')
    raster_data = 1000 * np.random.rand(512, 512).astype(np.uint16)

    profile = BASE_PROFILE.copy()
    profile.update(
        height=raster_data.shape[0],
        width=raster_data.shape[1],
        driver='PNG'
    )

    with rasterio.open(outfile, 'w', **profile) as dst:
        dst.write(raster_data, 1)

    assert not cog.validate(outfile)
